#include <osmocom/core/talloc.h>
#include <osmocom/core/linuxlist.h>

#include "v5x_internal.h"

static LLIST_HEAD(v5_instances);

struct v5x_instance *v5x_instance_alloc(void *ctx)
{
	struct v5x_instance *v5i = talloc_zero(ctx, struct v5x_instance);
	if (!v5i)
		return NULL;

	INIT_LLIST_HEAD(&v5i->interfaces);
	INIT_LLIST_HEAD(&v5i->user_ports);
	/* TODO: allocate ctrl_fi */

	llist_add_tail(&v5i->list, &v5_instances);

	return v5i;
}

static void v5x_ts_init(struct v5x_timeslot *v5ts, uint8_t nr, struct v5x_link *v5l, uint16_t l3_addr)
{
	v5ts->nr = nr;
	v5ts->link = v5l;
	v5ts->l3_address = l3_addr;
}

static void v5x_cchan_init(struct v5x_c_channel *v5cc, struct v5x_link *v5l, struct v5x_timeslot *v5ts, bool active)
{
	v5cc->link = v5l;
	v5cc->ts = v5ts;
	v5cc->active = active;
}

static void v5x_link_init(struct v5x_link *v5l, struct v5x_interface *v5if, uint8_t id)
{
	unsigned int i;

	v5l->id = id;
	v5l->interface = v5if;

	for (i = 1; i < ARRAY_SIZE(v5l->ts); i++) {
		v5x_ts_init(&v5l->ts[i], i, v5l, 0 /*l3_addr*/);
	}

	/* primary c-channel must always be present */
	v5x_cchan_init(&v5l->c_channel[0], v5l, &v5l->ts[16], true);
}

struct v5x_interface *v5x_interface_alloc(struct v5x_instance *v5i, enum v5x_dialect dialect)
{
	struct v5x_interface *v5if = talloc_zero(v5i, struct v5x_interface);
	//struct v5x_link *v5l;
	struct v5x_c_channel *primary_c_chan;

	v5if->instance = v5i;
	v5if->dialect = dialect;

	/* primary link must alwasy be present */
	v5x_link_init(&v5if->links[0], v5if, 0);
	v5if->primary_link = &v5if->links[0];

	primary_c_chan = &v5if->primary_link->c_channel[0];

	
	/* TODO: allocate fi */

	//lapd_dl_init2(&v5if->control.dl, k, v_rnage, maxf, "control");
	v5if->control.c_chan = primary_c_chan;

	//lapd_dl_init2(&v5if->pstn.dl, k, v_rnage, maxf, "pstn");
	v5if->pstn.c_chan = primary_c_chan;

	if (v5if->dialect == V5X_DIALECT_V52) {
		//lapd_dl_init2(&v5if->lcp.dl, k, v_rnage, maxf, "lcp");
		v5if->lcp.c_chan = primary_c_chan;

		//lapd_dl_init2(&v5if->bcc.dl, k, v_rnage, maxf, "bcc");
		v5if->bcc.c_chan = primary_c_chan;

		//lapd_dl_init2(&v5if->protection[0].dl, k, v_rnage, maxf, "protection0");
		v5if->protection[0].c_chan = primary_c_chan;

		//protection[1] ? 
	}

	llist_add_tail(&v5if->list, &v5i->interfaces);

	return v5if;
}
